
import sys

class Node:
    def __init__(self, id, available=True, local=False, services=None, cpu=0, client=None, ipList=None, hostname=None, config=None, instanceID=None, metadata=None, seq=0, ver=0, sender=None):
        self.id = id
        self.available = available
        self.local = local
        self.services = services if services is not None else []
        self.cpu = cpu
        self.client = client
        self.ipList = ipList if ipList is not None else []
        self.hostname = hostname
        self.config = config if config is not None else {}
        self.instanceID = instanceID
        self.metadata = metadata if metadata is not None else {}
        self.seq = seq
        self.ver = ver
        self.sender = sender

    def get_info(self):
        # return {k: v for k, v in self.__dict__.items() if k != 'id'} # when sending info, the transit should not send id and sender in the same payload
        return self.__dict__

class NodeCatalog:
    def __init__(self, registry=None, logger=None, node_id=None):
        self.nodes = {}
        self.registry = registry
        self.logger = logger
        self.node_id = node_id
        self.local_node = None
        self.ensure_local_node()


    def add_node(self, id, node: Node):
        self.nodes[id] = node
        if self.registry and hasattr(node, "services"):
            for service in node.services:
                actions = service.get("actions", {})
                for action_name in actions:
                    self.registry.add_action(action_name, id)
                events = service.get("events", {})
                for event in events:
                    self.registry.add_event(event, id)
        self.logger.info(f"Node \"{id}\" added.")

    def get_node(self, id):
        return self.nodes.get(id)

    def remove_node(self, id):
        if id in self.nodes:
            del self.nodes[id]


    def disconnect_node(self, id):
        node = self.get_node(id)
        if node:
            node.available = False
            self.logger.info(f"Node \"{id} \" is disconnected.")
            self.remove_node(id)

    def process_node_info(self, node_id, payload):
        node = self.get_node(node_id)
        if not node:
            node = Node(node_id)
            self.add_node(node)
        node.available = True
        node.cpu = payload.get("cpu", 0)
        node.services = payload.get("services", [])
        self.logger.info(f"Node \"{node_id}\" is connected.")


    def ensure_local_node(self):
        if not self.local_node:
            node = Node(self.node_id)
            self.local_node = node
            self.add_node(self.node_id, node)
        self.local_node.local = True
        self.local_node.client = {
            "type": "python",
            "langVersion": sys.version,
            # "version": version,
        }
        # TODO: local_node.services is essentialy different from the class Service
        # Debug print for service events
        for service in self.registry.__services__.values():
            for event in service.events():
                event_name = getattr(getattr(service, event), "_name", event)
                print(f"Service: {service.name}, Event: {event}, Name: {event_name}")

        self.local_node.services = [
                {
                    "name": service.name,
                    "fullName": service.name,
                    "settings": service.settings,
                    "metadata": service.metadata,
                    "actions": {
                        f"{service.name}.{action}": {
                            "rawName": action,
                            "name": f"{service.name}.{action}"
                        }
                        for action in service.actions()
                    },
                    "events": {
                        f"{getattr(getattr(service, event), "_name", event)}": {
                            "name": getattr(getattr(service, event), "_name", event)
                        }
                        for event in service.events()
                    }
                } 
                for service in self.registry.__services__.values() 
            ]



import datetime
import logging
import sys
import structlog

LOG_LEVEL_MAP = {
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG, 
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL
}

def get_parsed_log_level(log_level):
    return LOG_LEVEL_MAP.get(log_level, logging.INFO)

def moleculer_format_renderer(_, __, event_dict):
    timestamp = datetime.datetime.utcnow().isoformat(timespec="milliseconds") + "Z"
    level = event_dict.pop("level", "INFO").upper()
    node = event_dict.pop("node", "<unknown>")
    service = event_dict.pop("service", "<unspecified>")
    message = event_dict.pop("event", "")
    return f"[{timestamp}] {level:<5} {node}/{service}: {message}"


def get_logger(log_level, log_format="PLAIN"):
    log_level = get_parsed_log_level(log_level=log_level)
    logging.basicConfig(format="%(message)s", stream=sys.stdout, level=log_level)
    
    processors = [
        structlog.contextvars.merge_contextvars,
        structlog.processors.add_log_level,
    ]
    
    if log_format == "JSON":
        processors.extend([
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.format_exc_info,
            structlog.processors.JSONRenderer(),
        ])
    else:
        processors.append(moleculer_format_renderer)
    
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
        processors=processors,
    )
    
    return structlog.get_logger()

            



import asyncio
import signal
from pylecular.context import Context
from pylecular.discoverer import Discoverer
from pylecular.lifecycle import Lifecycle
from pylecular.node import NodeCatalog
from pylecular.registry import Registry 
from pylecular.settings import Settings
from pylecular.transit import Transit
from pylecular.logger import get_logger

class Broker:
    def __init__(self, 
                 id, 
                 settings: Settings=Settings(),
                 version: str="0.14.35",
                 namespace: str="default",
                 lifecycle=None,
                 registry=None,
                 node_catalog=None,
                 transit=None,
                 discoverer=None):
        self.id = id
        self.version = version
        self.namespace = namespace
        self.logger = get_logger(settings.log_level, settings.log_format).bind(
            node=self.id,
            service="BROKER",
            level=settings.log_level
        )
        self.lifecycle = lifecycle or Lifecycle(broker=self)
        self.registry = registry or Registry(node_id=self.id, logger=self.logger)
        self.node_catalog = node_catalog or NodeCatalog(logger=self.logger, node_id=self.id, registry=self.registry)
        self.transit = transit or Transit(settings=settings, node_id=self.id, registry=self.registry, node_catalog=self.node_catalog, lifecycle=self.lifecycle, logger=self.logger)
        self.discoverer = discoverer or Discoverer(broker=self)


    async def start(self):
        self.logger.info(f"Moleculer v{self.version} is starting...")
        self.logger.info(f"Namespace: {self.namespace}.")
        self.logger.info(f"Node ID: {self.id}.")
        self.logger.info(f"Transporter: {self.transit.transporter.name}.")
        await self.transit.connect()
        self.logger.info(f"✔ Service broker with {len(self.registry.__services__)} services started.")


    async def stop(self):
        await self.transit.disconnect()
        self.logger.info("Service broker is stopped. Good bye.")

    async def wait_for_shutdown(self):
        loop = asyncio.get_event_loop()
        shutdown_event = asyncio.Event()

        def signal_handler():
            shutdown_event.set()

        for sig in (signal.SIGINT, signal.SIGTERM):
            loop.add_signal_handler(sig, signal_handler)

        await shutdown_event.wait()
        await self.stop()

    async def wait_for_services(self, services=[]):
        while True:
            found = True
            for name in services:
                service = self.registry.get_service(name)
                if not service:
                    # check in remote nodes
                    for node in self.node_catalog.nodes.values():
                        if node.id != self.id:
                            for service_obj in node.services:
                                if service_obj.get("name") == name:
                                    service = service_obj
                                    break
                if not service:
                    found = False
                    break
            if found:
                return
            await asyncio.sleep(0.1)


    # TODO: fix service lyfecicle handling on catalog
    # TODO: if service is alive send INFO
    def register(self, service):
        self.registry.register(service)
        self.node_catalog.ensure_local_node()

    # TODO: support balancing strategies
    # TODO: support unbalanced
    async def call(self, action_name, params={}, meta={}):
        endpoint = self.registry.get_action(action_name)
        context = self.lifecycle.create_context(action=action_name, params=params, meta=meta)
        if endpoint and endpoint.is_local:
            return await endpoint.handler(context)
        elif endpoint and not endpoint.is_local:
            return await self.transit.request(endpoint, context)
        else:
            raise Exception(f"Action {action_name} not found.")
        

    async def emit(self, event_name, params={}, meta={}):
        endpoint = self.registry.get_event(event_name)
        context = self.lifecycle.create_context(event=event_name, params=params, meta=meta)
        if endpoint and endpoint.is_local:
            return await endpoint.handler(context)
        elif endpoint and not endpoint.is_local:
            return await self.transit.send_event(endpoint, context)
        
    
    async def broadcast(self, event_name, params={}, meta={}):
        endpoints = self.registry.get_all_events(event_name)
        context = self.lifecycle.create_context(event=event_name, params=params, meta=meta)
        await asyncio.gather(*[
            endpoint.handler(context) if endpoint.is_local 
            else self.transit.send_event(endpoint, context)
            for endpoint in endpoints
        ])


import uuid

class Context:
    # TODO: support stream
    def __init__(self, id, action=None, event=None, parent_id=None, params={}, meta={}, stream=False, broker=None):
        self.id = id
        self.action = action
        self.event = event
        self.params = params
        self.meta = meta
        self.parent_id = parent_id
        self.stream = stream
        self._broker = broker

    def unmarhshall(self):
        return {
            "id": self.id,
            "action": self.action,
            "event": self.event,
            "params": self.params,
            "meta": self.meta,
            "timeout": 0,
            "level": 1,
            "tracing": None,
            "parentID": self.parent_id,
            "stream": self.stream,
        }
    
    def marshall(self):
        return {
            "id": self.id,
            "action": self.action,
            "event": self.event,
            "params": self.params,
            "meta": self.meta,
            "timeout": 0,
            "level": 1,
            "tracing": None,
            "parentID": self.parent_id,
            "stream": self.stream,
        }

    async def _prepare_meta(self, meta={}):
        return {**self.meta, **meta}

    async def call(self, service_name, params={}, meta={}):
        meta = await self._prepare_meta(meta)
        return await self._broker.call(service_name, params, meta)

    async def emit(self, service_name, params={}, meta={}):
        meta = await self._prepare_meta(meta)
        return await self._broker.emit(service_name, params)

    async def broacast(self, service_name, params={}, meta={}):
        meta = await self._prepare_meta(meta)
        return await self._broker.broadcast(service_name, params)



from .base import Transporter
from pylecular.packet import Packet, Topics
import nats
import json

class NatsTransporter(Transporter):
    name = "nats"

    def __init__(self, connection_string, transit, handler=None, node_id=None):
        super().__init__(self.name)
        self.connection_string = connection_string
        self.transit = transit
        self.handler = handler
        self.node_id = node_id
        self.nc = None

    # TODO: maybe move it to base class
    # TODO: user real world serializer
    def _serialize(self, payload): 
        payload["ver"] = "4"
        payload["sender"] = self.node_id
        return json.dumps(payload).encode('utf-8')
    
    def get_topic_name(self, command, node_id=None):
        topic = f"MOL.{command}"
        if node_id:
            topic += f".{node_id}"
        return topic
    
    async def message_handler(self, msg):
        data = json.loads(msg.data.decode("utf-8"))
        type = Packet.from_topic(msg.subject)
        sender = data.get("sender")
        packet = Packet(type, sender, data)
        await self.handler(packet)

    async def publish(self, packet: Packet):
        topic = self.get_topic_name(packet.type.value, packet.target) 
        await self.nc.publish(topic, self._serialize(packet.payload))

    async def connect(self):
        self.nc = await nats.connect(self.connection_string)

    async def disconnect(self):
        if self.nc:
            await self.nc.close()
            self.nc = None

    async def subscribe(self, command, node_id=None):
        topic = self.get_topic_name(command, node_id)
        if self.handler is None:
            raise ValueError("Handler must be provided for subscription.")
        if not callable(self.message_handler) or not hasattr(self.message_handler, "__call__") or not hasattr(self.message_handler, "__code__") or not self.message_handler.__code__.co_flags & 0x80:
            raise ValueError("Handler must be an async function.")
        await self.nc.subscribe(topic, cb=self.message_handler)
    
    @classmethod
    def from_config(cls, config, transit, handler=None, node_id=None) -> "Transporter":
        return cls(connection_string=config["connection"], transit=transit, handler=handler, node_id=node_id)



import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # import pylecular
from pylecular.context import Context
from pylecular.service import Service
from pylecular.decorators import action, event
from pylecular.broker import Broker
    
class MySyservice(Service):
    name = "myService"

    def __init__(self):
        super().__init__(self.name)

    @event()
    async def checked(self, ctx: Context):
        print("checked ctx.id: " + ctx.id)
        await ctx.emit("done")
    
    @event()
    async def done(self, ctx: Context):
        print("done ctx.id: " + ctx.id)
        print("done called")

# Example usage
import asyncio

async def main():
    broker3 = Broker("broker-b3")
    await broker3.start()

    broker = Broker("broker-b1")
    mysvc = MySyservice()
    broker.register(mysvc)
    await broker.start()

    broker2 = Broker("broker-b2")
    mysvc2 = MySyservice()
    broker2.register(mysvc2)
    await broker2.start()


    await asyncio.sleep(1)

    await broker3.wait_for_services(["myService"])


    await broker3.broadcast("checked", {})

    await broker3.wait_for_shutdown()

asyncio.run(main())


import pytest
import asyncio
import signal
import os
from unittest.mock import Mock, AsyncMock
from pylecular.broker import Broker
from pylecular.settings import Settings
from pylecular.service import Service
from pylecular.decorators import action, event
from pylecular.transit import Transit
from pylecular.registry import Registry
from pylecular.node import NodeCatalog
from pylecular.lifecycle import Lifecycle

class TestService(Service):
    def __init__(self):
        super().__init__(name="test")

    @action()
    async def hello(self, _):
        return "Hello!"

    @event()
    async def test_event(self, _):
        return "Event received"

    @event(name="custom.event")
    async def another_event(self, _):
        return "Custom event received"
    

@pytest.fixture(scope="function")
def mock_transit():
    transit = AsyncMock(spec=Transit)
    transit.connect = AsyncMock()
    transit.disconnect = AsyncMock()
    transit.transporter = Mock(name="mock_nats")
    return transit

@pytest.fixture
def mock_registry():
    registry = Mock(spec=Registry)
    registry.__services__ = {}
    return registry

@pytest.fixture
def mock_node_catalog():
    catalog = Mock(spec=NodeCatalog)
    catalog.nodes = Mock()
    return catalog

@pytest.fixture
def mock_lifecycle():
    return Mock(spec=Lifecycle)

@pytest.fixture
async def broker(mock_transit, mock_registry, mock_node_catalog, mock_lifecycle):
    # Create and set an event loop for the test
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    
    settings = Settings(transporter="mock://localhost:4222")
    broker = Broker(
        "test-node",
        settings=settings,
        transit=mock_transit,
        registry=mock_registry,
        node_catalog=mock_node_catalog,
        lifecycle=mock_lifecycle
    )
    return broker

@pytest.mark.asyncio
async def test_broker_initialization(broker):
    broker_instance = await broker
    assert broker_instance.id == "test-node"
    assert broker_instance.version == "0.14.35"
    assert broker_instance.namespace == "default"
    assert broker_instance.registry is not None
    assert broker_instance.transit is not None
    

@pytest.mark.asyncio
async def test_broker_start(broker, mock_transit):
    broker_instance = await broker
    await broker_instance.start()


@pytest.mark.asyncio
async def test_broker_stop(broker, mock_transit):
    broker_instance = await broker
    await broker_instance.stop()


@pytest.mark.asyncio
async def test_broker_stop(broker, mock_transit):
    broker_instance = await broker
    await broker_instance.stop()
    mock_transit.disconnect.assert_called_once()

@pytest.mark.asyncio
async def test_broker_register_service(broker, mock_registry, mock_node_catalog):
    broker_instance = await broker
    service = TestService()
    broker_instance.register(service)
    mock_registry.register.assert_called_once_with(service)
    mock_node_catalog.ensure_local_node.assert_called_once()


@pytest.mark.asyncio
async def test_broker_call_local_action(broker, mock_registry, mock_lifecycle):
    endpoint = Mock(is_local=True)
    endpoint.handler = AsyncMock(return_value="result")
    mock_registry.get_action.return_value = endpoint
    
    context = Mock()
    mock_lifecycle.create_context.return_value = context
    
    broker_instance = await broker

    result = await broker_instance.call("test.hello", {"param": "value"})
    
    assert result == "result"
    mock_registry.get_action.assert_called_once_with("test.hello")
    endpoint.handler.assert_called_once_with(context)


@pytest.mark.asyncio
async def test_broker_call_remote_action(broker, mock_registry, mock_transit, mock_lifecycle):
    endpoint = Mock(is_local=False, node_id="remote-node")
    mock_registry.get_action.return_value = endpoint
    mock_transit.request.return_value = "remote result"
    
    context = Mock()
    mock_lifecycle.create_context.return_value = context
    
    broker_instance = await broker

    result = await broker_instance.call("remote.action")
    
    assert result == "remote result"
    mock_transit.request.assert_called_once_with(endpoint, context)


@pytest.mark.asyncio
async def test_broker_call_nonexistent_action(broker, mock_registry):
    broker_instance = await broker

    mock_registry.get_action.return_value = None
    with pytest.raises(Exception, match="Action nonexistent.action not found."):
        await broker_instance.call("nonexistent.action")

@pytest.mark.asyncio
async def test_broker_emit_local_event(broker, mock_registry, mock_lifecycle):
    endpoint = Mock(is_local=True)
    endpoint.handler = AsyncMock()
    mock_registry.get_event.return_value = endpoint
    
    context = Mock()
    mock_lifecycle.create_context.return_value = context
    
    broker_instance = await broker

    await broker_instance.emit("test_event", {"param": "value"})
    
    endpoint.handler.assert_called_once_with(context)


@pytest.mark.asyncio
async def test_broker_emit_remote_event(broker, mock_registry, mock_transit, mock_lifecycle):
    endpoint = Mock(is_local=False)
    mock_registry.get_event.return_value = endpoint
    
    context = Mock()
    mock_lifecycle.create_context.return_value = context

    broker_instance = await broker
    
    await broker_instance.emit("remote.event")
    
    mock_transit.send_event.assert_called_once_with(endpoint, context)

@pytest.mark.asyncio
async def test_broker_broadcast_event(broker, mock_registry, mock_transit, mock_lifecycle):
    local_endpoint = Mock(is_local=True)
    local_endpoint.handler = AsyncMock()
    remote_endpoint = Mock(is_local=False)
    
    mock_registry.get_all_events.return_value = [local_endpoint, remote_endpoint]
    
    context = Mock()
    mock_lifecycle.create_context.return_value = context
    
    broker_instance = await broker

    await broker_instance.broadcast("test_event")
    
    local_endpoint.handler.assert_called_once_with(context)
    mock_transit.send_event.assert_called_once_with(remote_endpoint, context)

@pytest.mark.asyncio
async def test_wait_for_services_found_locally(broker, mock_registry):
    mock_registry.get_service.return_value = Mock()
    broker_instance = await broker
    await broker_instance.wait_for_services(["test"])
    mock_registry.get_service.assert_called_once_with("test")

@pytest.mark.asyncio
async def test_wait_for_services_found_remotely(broker, mock_registry, mock_node_catalog):
    mock_registry.get_service.return_value = None
    remote_node = Mock()
    remote_node.id = "remote-node"
    remote_node.services = [{"name": "test"}]
    mock_node_catalog.nodes.values.return_value = [remote_node]
    broker_instance = await broker
    await broker_instance.wait_for_services(["test"])



import uuid
from pylecular.context import Context


# TODO: handle context list and graceful shutdown
class Lifecycle:

    def __init__(self, broker):
        self.broker = broker

    def create_context(self, id=uuid.uuid4(), event=None, action=None, parent_id=None, params={}, meta={}, stream=False):
        return Context(
            str(id),
            action=action,
            event=event,
            parent_id=parent_id,
            params=params,
            meta=meta,
            stream=stream,
            broker=self.broker
        )
    

    def rebuild_context(self, context_dict):
        return Context(
            str(context_dict.get('id')),
            action=context_dict.get('action'),
            event=context_dict.get('event'),
            parent_id=context_dict.get('parent_id'),
            params=context_dict.get('params', {}),
            meta=context_dict.get('meta', {}),
            stream=context_dict.get('stream', False),
            broker=self.broker
        )


import asyncio


class Discoverer:
    def __init__(self, broker):
        self.broker = broker
        self.transit = broker.transit
        self.__setup_timers__()

    def __setup_timers__(self):
        async def periodic_beat():
            while True:
                await asyncio.sleep(5)
                await self.transit.beat()
        asyncio.create_task(periodic_beat())

    
    # TODO: kill timers



from abc import ABC, abstractmethod
import importlib

class Transporter(ABC):
    def __init__(self, name):
        self.name = name
        
    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def disconnect(self):
        pass

    @abstractmethod
    async def publish(self):
        pass

    @abstractmethod
    async def subscribe(self, command, topic):
        pass

    @classmethod
    def get_by_name(cls, name: str, config: dict, transit, handler=None, node_id=None) -> "Transporter":
        importlib.import_module("pylecular.transporter.nats")

        for subclass in cls.__subclasses__():
            if subclass.__name__.lower().startswith(name.lower()):
                return subclass.from_config(config, transit, handler, node_id)
        raise ValueError(f"No transporter found for: {name}")

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict, transit, handler=None, node_id=None) -> "Transporter":
        pass




import pytest
from pylecular.service import Service
from pylecular.decorators import action, event


class MyService(Service):
    name = "myService"

    @action()
    def say_hello(self):
        return "hello"
    
    @event()
    def receive(self, ctx):
        return f"Received: {ctx}"
    
    def __internal(self):
        return "internal"


def test_service_init():
    s = Service("auth", settings={"foo": "bar"})
    assert s.name == "auth"
    assert s.settings == {"foo": "bar"}
    assert s.metadata == {}

def test_service_default_settings():
    s = Service("test")
    assert s.settings == {}

def test_actions_returns_callable_attrs():
    s = MyService("test")
    actions = s.actions()
    assert "say_hello" in actions
    assert "__internal" not in actions  # because it doesn't start with __
    assert "actions" not in actions
    assert "events" not in actions

def test_events_returns_callable_attrs():
    s = MyService("test")
    events = s.events()
    assert "receive" in events
    assert "say_hello" not in events
    assert "__internal" not in events  # because it doesn't start with __
    assert "actions" not in events
    assert "events" not in events




import pytest
from unittest.mock import Mock, AsyncMock
from pylecular.context import Context

@pytest.fixture
def mock_broker():
    broker = Mock()
    broker.call = AsyncMock(return_value="call_result")
    broker.emit = AsyncMock(return_value="emit_result")
    broker.broadcast = AsyncMock(return_value="broadcast_result")
    return broker

@pytest.fixture
def context(mock_broker):
    return Context(
        id="test-id",
        action="test.action",
        event="test.event",
        parent_id="parent-id",
        params={"key": "value"},
        meta={"meta_key": "meta_value"},
        stream=False,
        broker=mock_broker
    )

def test_context_initialization(context):
    assert context.id == "test-id"
    assert context.action == "test.action"
    assert context.event == "test.event"
    assert context.parent_id == "parent-id"
    assert context.params == {"key": "value"}
    assert context.meta == {"meta_key": "meta_value"}
    assert context.stream is False
    assert context._broker is not None

def test_context_initialization_with_defaults():
    ctx = Context(id="test-id")
    assert ctx.id == "test-id"
    assert ctx.action is None
    assert ctx.event is None
    assert ctx.parent_id is None
    assert ctx.params == {}
    assert ctx.meta == {}
    assert ctx.stream is False
    assert ctx._broker is None

def test_marshall(context):
    marshalled = context.marshall()
    assert marshalled == {
        "id": "test-id",
        "action": "test.action",
        "event": "test.event",
        "params": {"key": "value"},
        "meta": {"meta_key": "meta_value"},
        "timeout": 0,
        "level": 1,
        "tracing": None,
        "parentID": "parent-id",
        "stream": False
    }

def test_unmarshall(context):
    unmarshalled = context.unmarhshall()
    assert unmarshalled == {
        "id": "test-id",
        "action": "test.action",
        "event": "test.event",
        "params": {"key": "value"},
        "meta": {"meta_key": "meta_value"},
        "timeout": 0,
        "level": 1,
        "tracing": None,
        "parentID": "parent-id",
        "stream": False
    }

@pytest.mark.asyncio
async def test_prepare_meta(context):
    new_meta = {"new_key": "new_value"}
    result = await context._prepare_meta(new_meta)
    assert result == {
        "meta_key": "meta_value",
        "new_key": "new_value"
    }

@pytest.mark.asyncio
async def test_prepare_meta_empty(context):
    result = await context._prepare_meta()
    assert result == {"meta_key": "meta_value"}

@pytest.mark.asyncio
async def test_call(context, mock_broker):
    params = {"param_key": "param_value"}
    meta = {"meta_key": "new_meta_value"}
    
    result = await context.call("service.action", params, meta)
    
    assert result == "call_result"
    mock_broker.call.assert_called_once_with(
        "service.action",
        {"param_key": "param_value"},
        {"meta_key": "new_meta_value"}
    )

@pytest.mark.asyncio
async def test_emit(context, mock_broker):
    params = {"param_key": "param_value"}
    meta = {"meta_key": "new_meta_value"}
    
    result = await context.emit("service.event", params, meta)
    
    assert result == "emit_result"
    mock_broker.emit.assert_called_once_with(
        "service.event",
        {"param_key": "param_value"}
    )

@pytest.mark.asyncio
async def test_broadcast(context, mock_broker):
    params = {"param_key": "param_value"}
    meta = {"meta_key": "new_meta_value"}
    
    result = await context.broacast("service.event", params, meta)
    
    assert result == "broadcast_result"
    mock_broker.broadcast.assert_called_once_with(
        "service.event",
        {"param_key": "param_value"}
    )

@pytest.mark.asyncio
async def test_call_without_broker(context):
    context._broker = None
    with pytest.raises(AttributeError):
        await context.call("service.action")

@pytest.mark.asyncio
async def test_emit_without_broker(context):
    context._broker = None
    with pytest.raises(AttributeError):
        await context.emit("service.event")

@pytest.mark.asyncio
async def test_broadcast_without_broker(context):
    context._broker = None
    with pytest.raises(AttributeError):
        await context.broacast("service.event")



from pylecular.node import NodeCatalog


class Action:
    def __init__(self, name, node_id, is_local, handler=None):
        self.name = name
        self.handler = handler
        self.node_id = node_id
        self.is_local = is_local


class Event:
    def __init__(self, name, node_id, is_local=False, handler=None):
        self.name = name
        self.node_id = node_id
        self.handler = handler
        self.is_local = is_local



class Registry:
    def __init__(self, node_id=None, logger=None):
        self.__services__ = {} # local services
        self.__actions__ = []
        self.__events__ = []
        self.__node_id__ = node_id
        self.__logger__ = logger

    # TODO: handle service removal
    # TODO: handle remove action removal

    def register(self, service):
        self.__services__[service.name] = service
        self.__actions__.extend([
            Action(f"{service.name}.{action}",self.__node_id__, is_local=True, handler=getattr(service, action))
            for action in service.actions()
        ])
        for event in service.events():
            handler = getattr(service, event)
            is_callable = callable(handler)
            event_name = getattr(handler, "_name", event) if is_callable else event
            print(f"Event {event}: callable={is_callable}, name={event_name}")
        self.__events__.extend([
            Event(getattr(getattr(service, event), "_name", event), self.__node_id__, is_local=True, handler=getattr(service, event))
            for event in service.events()
        ])
        for event in self.__events__:
            print(f"Event {event.name} from node {event.node_id} (local={event.is_local})")
        # self.logger.info(f"Service {service.name} registered with {len(self.actions)} actions and {len(self.events)} events.")

    def get_service(self, name):
        return self.__services__.get(name)
    
    def add_action(self, name, node_id):
        action = Action(name, node_id, is_local=False)
        self.__actions__.append(action)
        
    def add_event(self, name, node_id):
        event = Event(name, node_id, is_local=False)
        self.__events__.append(event)

    def get_action(self, name) -> Action:
        action = [a for a in self.__actions__ if a.name == name]
        if action:
            return action[0]

    def get_all_events(self, name):
        return [a for a in self.__events__ if a.name == name]

    def get_event(self, name) -> Event:
        event = self.get_all_events(name)
        if event:
            return event[0]



def action(params=None):
    def decorator(func):
        func._is_action = True
        return func
    return decorator

def event(name=None,params=None):
    def decorator(func):
        func._is_event = True
        func._name = name if name else func.__name__
        return func
    return decorator



class Settings:
    def __init__(self, transporter="nats://localhost:4222", serializer="JSON", log_level="INFO", log_format="PLAIN"):
        self.transporter = transporter
        self.serializer = serializer
        self.log_level = log_level
        self.log_format = log_format



import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # import pylecular
from pylecular.context import Context
from pylecular.service import Service
from pylecular.decorators import action, event
from pylecular.broker import Broker
    
class MySyservice(Service):
    name = "myService"

    def __init__(self):
        super().__init__(self.name)

    # TODO: add validation
    @action(params=["param1", "param2"])
    async def foo(self, ctx: Context):
        return  100


    @action(params=[])
    async def bar(self, ctx: Context):
        a = await ctx.call("myService.foo", {}) # internal context call
        res = await ctx.call("math.add", { "a": a, "b": 50 }) # remote service call
        return res
    
    @event()
    async def checked(self, ctx: Context):
        print("checked called")

    @event(name="foo.bar.lol")
    async def complex_handler(self, ctx: Context):
        print("compled called")
    
    
# Example usage
import asyncio

async def main():
    broker = Broker("broker1")

    mysvc = MySyservice()

    broker.register(mysvc)

    await broker.start()

    res = await broker.call("myService.foo", {"param1": "aaaa"})

    broker.logger.info(f"local res is {res}")

    broker.logger.info("waiting for services")

    await broker.wait_for_services(["math"])

    broker.logger.info("remote services ready")

    mathRes = await broker.call("math.add", {"a": 1, "b": 50 })  # remote action on tests/intetgration

    broker.logger.info(f"math res is {mathRes}")

    barREs = await broker.call("myService.bar", {}, meta={ "tenant": "foobar-company" })

    broker.logger.info(f"bar res is {barREs}")

    await broker.emit("checked", {})

    await broker.wait_for_shutdown()

asyncio.run(main())



import asyncio
from pylecular.context import Context
from pylecular.node import Node
from pylecular.transporter.base import Transporter
from pylecular.packet import Packet,Topics
from pylecular.transporter.base import Transporter
import psutil


class Transit:
    def __init__(self, node_id=None, registry=None, node_catalog=None, settings=None, logger=None, lifecycle=None):
        self.node_id = node_id
        self.registry = registry
        self.node_catalog = node_catalog
        self.logger = logger
        self.transporter = Transporter.get_by_name(
            settings.transporter.split("://")[0], 
            {"connection": settings.transporter}, 
            transit=self, handler=self.__message_handler__, node_id=node_id)
        self._pending_requests = {} 
        self.lifecycle = lifecycle


    async def __message_handler__(self, packet: Packet):
        if packet.type == Topics.INFO:
            await self.info_handler(packet)
        elif packet.type == Topics.DISCOVER:
            await self.discover_handler(packet)
        elif packet.type == Topics.HEARTBEAT:
            await self.heartbeat_handler(packet)
        elif packet.type == Topics.REQUEST:
            await self.request_handler(packet)
        elif packet.type == Topics.RESPONSE:
            await self.response_handler(packet)
        elif packet.type == Topics.EVENT:
            await self.event_handler(packet)
        elif packet.type == Topics.DISCONNECT:
            await self.disconnect_handler(packet)

    async def __make_subscriptions__(self):
        await self.transporter.subscribe(Topics.INFO.value)
        await self.transporter.subscribe(Topics.INFO.value, self.node_id)
        await self.transporter.subscribe(Topics.DISCONNECT.value)
        await self.transporter.subscribe(Topics.HEARTBEAT.value)
        await self.transporter.subscribe(Topics.REQUEST.value, self.node_id)
        await self.transporter.subscribe(Topics.RESPONSE.value, self.node_id)
        await self.transporter.subscribe(Topics.EVENT.value, self.node_id)


    async def connect(self):
        await self.transporter.connect()
        await self.discover()
        await self.send_node_info()
        await self.__make_subscriptions__()

        
    async def disconnect(self):
        await self.publish(Packet(Topics.DISCONNECT, None, {}))
        for future in self._pending_requests.values():
            if not future.done():
                future.cancel()
        self._pending_requests.clear()
        await self.transporter.disconnect()

    async def publish(self, packet: Packet):
        await self.transporter.publish(packet)


    async def discover(self):
        await self.publish(Packet(Topics.DISCOVER, None, {}))


    async def beat(self):
        heartbeat = { # TODO: move to node catalog
            "cpu": psutil.cpu_percent(interval=1),

        }
        await self.publish(Packet(Topics.HEARTBEAT, None, heartbeat))

    async def send_node_info(self):
        await self.publish(Packet(Topics.INFO, None, self.node_catalog.local_node.get_info()))


    async def discover_handler(self, packet: Packet):
        await self.send_node_info()

    async def heartbeat_handler(self, packet: Packet):
        # TODO: setup heartbeat times
        pass

    async def info_handler(self, packet: Packet):
        node = Node(
            id=packet.payload.get("id"),
            **{k: v for k, v in packet.payload.items() if k != "id"})
        self.node_catalog.add_node(packet.target, node)

    async def disconnect_handler(self, packet: Packet):
        self.node_catalog.disconnect_node(packet.target)

    async def event_handler(self,packet: Packet):
        endpoint = self.registry.get_event(packet.payload.get("event"))
        if endpoint and endpoint.is_local:
            context = self.lifecycle.rebuild_context(packet.payload)
            try:
                await endpoint.handler(context)
            except Exception as e:
                self.logger.error(f"Failed to process event {endpoint.name}")
                self.logger.error(e)

    async def request_handler(self, packet: Packet):
        endpoint = self.registry.get_action(packet.payload.get("action"))
        if endpoint and endpoint.is_local:
            context = self.lifecycle.rebuild_context(packet.payload)
            try:
                result = await endpoint.handler(context)
                response = {
                    "id": context.id,
                    "data": result,
                    "success": True,
                    "meta": {}
                }
                await self.publish(Packet(Topics.RESPONSE, packet.target, response))
            except Exception as e:
                self.logger.error(f"Failed call to {endpoint.name}.", e)

    async def response_handler(self, packet: Packet):
        req_id = packet.payload.get("id")
        future = self._pending_requests.pop(req_id, None)
        if future:
            future.set_result(packet.payload)

    async def request(self, endpoint, context):
        req_id = context.id
        future = asyncio.get_running_loop().create_future()
        self._pending_requests[req_id] = future
        await self.publish(Packet(Topics.REQUEST, endpoint.node_id, context.marshall()))
        try:
            response = await asyncio.wait_for(future, 5000) # TODO: fetch timeout from settings
            return response.get("data")
        except asyncio.TimeoutError:
            self._pending_requests.pop(req_id, None)
            # Optionally log or raise a custom exception
            return None
        

    async def send_event(self, endpoint, context):
        await self.publish(Packet(Topics.EVENT, endpoint.node_id, context.marshall()))





from enum import Enum
import string

class Topics(Enum):
    HEARTBEAT = "HEARTBEAT"
    EVENT = "EVENT"
    DISCONNECT = "DISCONNECT"
    DISCOVER = "DISCOVER"
    INFO = "INFO"
    REQUEST = "REQ"
    RESPONSE = "RES"


class Packet:
    def __init__(self, type: string, target, payload):
        self.type = type
        self.target = target
        self.payload = payload


    @staticmethod
    def from_topic(topic: string):
        parts = topic.split(".")
        return Topics(parts[1]) # TODO: ensure and test



import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) # import pylecular
from pylecular.context import Context
from pylecular.service import Service
from pylecular.decorators import action, event
from pylecular.broker import Broker
from sklearn.linear_model import LinearRegression # require scikit-learn
import numpy as np
import numpy as np

class MLService(Service):
    name = "ml"

    def __init__(self):
        super().__init__(self.name)
        X = np.array([[1], [2], [3], [4]])
        y = np.array([2, 4, 6, 8])
        self.model = LinearRegression()
        self.model.fit(X, y)

    @action(params=["x"])
    async def predict(self, ctx: Context):
        x = float(ctx.params.get("x"))
        prediction = self.model.predict([[x]])
        return float(prediction[0])

# Example usage
import asyncio

async def main():
    broker = Broker("broker1")

    broker.register(MLService())

    await broker.start()

    res = await broker.call("ml.predict", {"x": 4 })

    broker.logger.info(f"ml predicted {res}")


asyncio.run(main())




class Service:
    def __init__(self, name, settings=None):
        self.name = name # support version as well
        self.settings = settings or {}
        self.metadata = {}

    def actions(self):
        return [
            attr for attr in dir(self)
            if callable(getattr(self, attr)) and getattr(getattr(self, attr), "_is_action", False)
        ]
    
    def events(self):
        return [
            attr for attr in dir(self)
            if callable(getattr(self, attr)) and getattr(getattr(self, attr), "_is_event", False)
        ]

